from tortoise.models import Model
from tortoise import fields
from .feed_entry import FeedEntry


class Feed(Model):
    id = fields.IntField(primary_key=True)
    name = fields.TextField()
    last_update = fields.DatetimeField(auto_now=True)
    entries: fields.ReverseRelation[FeedEntry]



import asyncio
from tortoise.models import Model
from tortoise import fields
from tortoise.signals import post_delete
from typing import Type
from anyio import Path
from telegram_to_rss.config import static_path


class FeedEntry(Model):
    id = fields.TextField(primary_key=True)
    feed = fields.ForeignKeyField(
        "models.Feed", on_delete=fields.CASCADE, related_name="entries"
    )
    message = fields.TextField()
    date = fields.DatetimeField()
    media = fields.JSONField(default=[])
    has_unsupported_media = fields.BooleanField(default=False)


@post_delete
async def remove_associated_file(
    sender: Type[FeedEntry],
    instance: FeedEntry,
) -> None:
    await asyncio.gather(
        *[
            Path(static_path).joinpath(media_relative_path).unlink(missing_ok=True)
            for media_relative_path in instance.media
        ]
    )



from telegram_to_rss.server import app
from telegram_to_rss.config import bind
import asyncio
from hypercorn.config import Config
from hypercorn.asyncio import serve
import argparse


# https://stackoverflow.com/a/46877092
def parse_hostport(bind_str: str | None) -> tuple[str | None, int | None]:
    if bind_str is None:
        return (None, None)

    out = bind_str.rsplit(":", 1)
    try:
        out[1] = int(out[1])
    except (IndexError, ValueError):
        # couldn't parse the last component as a port, so let's
        # assume there isn't a port.
        out = (bind_str, None)
    return out


def main():
    parser = argparse.ArgumentParser(
        prog="Telegram to RSS",
        description="Generate an RSS feed from your Telegram chats",
    )
    parser.add_argument("-d", "--dev", action="store_true")
    args = parser.parse_args()

    if args.dev:
        [host, port] = parse_hostport(bind)
        app.run(debug=True, host=host, port=port)
    else:
        config = Config()
        if bind:
            config.bind = bind
        asyncio.run(serve(app, config))



import asyncio
from quart import Quart, render_template
from telegram_to_rss.client import TelegramToRssClient
from telegram_to_rss.config import (
    api_hash,
    api_id,
    session_path,
    password,
    static_path,
    feed_size_limit,
    initial_feed_size,
    update_interval_seconds,
    db_path,
    loglevel,
)
from telegram_to_rss.qr_code import get_qr_code_image
from telegram_to_rss.db import init_feeds_db, close_feeds_db
from telegram_to_rss.generate_feed import update_feeds_cache
from telegram_to_rss.poll_telegram import TelegramPoller, update_feeds_in_db
from telegram_to_rss.models import Feed
import logging

logging.basicConfig(level=loglevel)

app = Quart(__name__, static_folder=static_path, static_url_path="/static")
client = TelegramToRssClient(
    session_path=session_path, api_id=api_id, api_hash=api_hash, password=password
)
telegram_poller = TelegramPoller(
    client=client,
    message_limit=feed_size_limit,
    new_feed_limit=initial_feed_size,
    static_path=static_path,
)
rss_task: asyncio.Task | None = None


async def start_rss_generation():
    global rss_task

    logging.info("start_rss_generation")

    async def update_rss():
        global rss_task

        try:
            logging.info("update_rss -> db")
            await update_feeds_in_db(telegram_poller=telegram_poller)

            logging.info("update_rss -> cache")
            await update_feeds_cache(feed_render_dir=static_path)

            logging.info("update_rss -> sleep")
            await asyncio.sleep(update_interval_seconds)

            logging.info("update_rss -> scheduling a new run")
            loop = asyncio.get_event_loop()
            rss_task = loop.create_task(update_rss())
        except Exception as e:
            rss_task = None
            logging.error(f"update_rss -> error: {e}")
            raise e

    await client.start()

    loop = asyncio.get_event_loop()
    rss_task = loop.create_task(update_rss())

    logging.info("start_rss_generation -> done")


@app.before_serving
async def startup():
    global rss_task

    logging.info("startup")

    await init_feeds_db(db_path=db_path)
    loop = asyncio.get_event_loop()
    rss_task = loop.create_task(start_rss_generation())

    logging.info("startup -> done")


@app.after_serving
async def cleanup():
    logging.info("cleanup")

    if rss_task is not None:
        rss_task.cancel()
    await client.stop()
    await close_feeds_db()

    logging.info("cleanup -> done")


@app.route("/")
async def root():
    logging.debug("GET /root %s", bool(client.qr_code_url))

    if client.qr_code_url is not None:
        qr_code_image = get_qr_code_image(client.qr_code_url)
        return await render_template("qr_code.html", qr_code=qr_code_image)

    feeds = await Feed.all()
    logging.debug("GET /root -> feeds %s", len(feeds))

    return await render_template("feeds.html", user=client.user, feeds=feeds)



from pathlib import Path
from quart import utils
from telegram_to_rss.models import Feed, FeedEntry
from tortoise.query_utils import Prefetch
from telegram_to_rss.config import base_url
from telegram_to_rss.poll_telegram import parse_feed_entry_id
import re
from telegram_to_rss.client import telethon_dialog_id_to_tg_id
import xml.etree.ElementTree as ET
import logging

CLEAN_TITLE = re.compile("<.*?>")


def clean_title(raw_html):
    cleantext = re.sub(CLEAN_TITLE, "", raw_html).replace("\n", " ").strip()
    return cleantext


def generate_feed(feed_render_dir: Path, feed: Feed):
    logging.info("generate_feed %s %s", feed.name, feed.id)

    feed_url = "https://t.me/c/{}".format(telethon_dialog_id_to_tg_id(feed.id))

    rss_root_el = ET.Element("rss", {"version": "2.0"})

    rss_feed_el = ET.SubElement(rss_root_el, "channel")

    ET.SubElement(rss_feed_el, "title").text = feed.name
    ET.SubElement(rss_feed_el, "pubDate").text = feed.last_update.isoformat()
    ET.SubElement(
        rss_feed_el,
        "link",
        {"href": feed_url},
    )
    ET.SubElement(rss_feed_el, "description").text = feed.name

    for feed_entry in feed.entries:
        [feed_id, entry_id] = parse_feed_entry_id(feed_entry.id)
        feed_entry_url = "https://t.me/c/{}/{}".format(
            telethon_dialog_id_to_tg_id(feed_id), entry_id
        )

        rss_item_el = ET.SubElement(rss_feed_el, "item")

        ET.SubElement(rss_item_el, "guid").text = feed_entry_url

        message_text = clean_title(feed_entry.message)
        title = message_text[:100]
        ET.SubElement(rss_item_el, "title").text = title

        images = ""
        for media_path in feed_entry.media:
            media_url = "{}/static/{}".format(base_url, media_path)
            images += '<br /><img src="{}"/>'.format(media_url)
        content = feed_entry.message.replace("\n", "<br />") + images
        if feed_entry.has_unsupported_media:
            content += "<br /><strong>This message has unsupported attachment. Open Telegram to view it.</strong>"
        ET.SubElement(rss_item_el, "description").text = content

        ET.SubElement(rss_item_el, "pubDate").text = feed_entry.date.isoformat()
        ET.SubElement(rss_item_el, "link", {"href": feed_entry_url}).text = (
            feed_entry_url
        )

    final_feed_file = feed_render_dir.joinpath("{}.xml".format(feed.id))

    rss_xml_tree = ET.ElementTree(rss_root_el)
    rss_xml_tree.write(
        file_or_filename=final_feed_file, encoding="UTF-8", short_empty_elements=True
    )

    logging.info("generate_feed -> done %s %s", feed.name, feed.id)


async def update_feeds_cache(feed_render_dir: str):
    feeds = await Feed.all().prefetch_related(
        Prefetch("entries", queryset=FeedEntry.all().order_by("-date"))
    )

    for feed in feeds:
        await utils.run_sync(generate_feed)(feed_render_dir, feed)



from tortoise import Tortoise


async def init_feeds_db(db_path: str):
    await Tortoise.init(
        db_url="sqlite://{}".format(db_path),
        modules={"models": ["telegram_to_rss.models"]},
    )
    # Generate the schema
    await Tortoise.generate_schemas(safe=True)


async def close_feeds_db():
    await Tortoise.close_connections()



import qrcode
from io import BytesIO
from base64 import b64encode


def get_qr_code_image(qr_code: str):
    buffer = BytesIO()
    img = qrcode.make(qr_code)
    img.save(buffer)
    encoded_img = b64encode(buffer.getvalue()).decode()

    data_uri = "data:image/png;base64,{}".format(encoded_img)
    return data_uri



import os
from pathlib import Path
from platformdirs import user_data_dir

api_id = int(os.environ.get("TG_API_ID"))
api_hash = os.environ.get("TG_API_HASH")
password = os.environ.get("TG_PASSWORD")

update_interval_seconds = int(os.environ.get("UPDATE_INTERVAL") or 3600)
feed_size_limit = int(os.environ.get("FEED_SIZE") or 200)
initial_feed_size = int(os.environ.get("INITIAL_FEED_SIZE") or 50)
base_url = os.environ.get("BASE_URL")
bind = os.environ.get("BIND") or "127.0.0.1:3042"

loglevel = os.environ.get("LOGLEVEL", "INFO").upper()

data_dir = (
    Path(os.environ.get("DATA_DIR"))
    if os.environ.get("DATA_DIR")
    else Path(user_data_dir()).joinpath("telegram_to_rss")
)
session_path = data_dir.joinpath("telegram_to-rss.session")
static_path = data_dir.joinpath("static")
db_path = data_dir.joinpath("feeds.db")

data_dir.mkdir(mode=0o700, parents=True, exist_ok=True)
static_path.mkdir(mode=0o700, exist_ok=True)



from telethon import TelegramClient, types, errors, custom
from telegram_to_rss.consts import TELEGRAM_NOTIFICATIONS_DIALOG_ID
from telethon.utils import resolve_id
from telegram_to_rss.consts import MESSAGE_FETCH_HARD_LIMIT
import logging


class TelegramToRssClient:
    _telethon: TelegramClient
    _qr_code_url: str | None = None
    _user: types.User = None
    _password: str | None = None

    def __init__(
        self, session_path: str, api_id: int, api_hash: str, password: str | None = None
    ):
        self._telethon = TelegramClient(
            session=session_path, api_id=api_id, api_hash=api_hash
        )
        self._telethon.parse_mode = "html"
        self._password = password

    async def start(self):
        await self._telethon.connect()
        is_authorized = await self._telethon.is_user_authorized()

        if not is_authorized:
            try:
                qr_login_req = await self._telethon.qr_login()
                self._qr_code_url = qr_login_req.url
                await qr_login_req.wait()
            except errors.SessionPasswordNeededError:
                if self._password is None:
                    raise Exception(
                        "2FA enabled and requires a password, but no password is provided."
                    )
                await self._telethon.sign_in(password=self._password)

        self._qr_code_url = None
        self._user = await self._telethon.get_me()

    async def stop(self):
        if self._telethon.is_connected():
            await self._telethon.disconnect()

    async def list_dialogs(self) -> list[custom.Dialog]:
        all_dialogs = await self._telethon.get_dialogs()
        filtered_dialogs = [
            dialog
            for dialog in all_dialogs
            if (
                dialog.id != TELEGRAM_NOTIFICATIONS_DIALOG_ID
                and dialog.entity.id != self._user.id
            )
        ]
        return filtered_dialogs

    async def get_dialog_messages(
        self,
        dialog: custom.Dialog,
        limit: int = MESSAGE_FETCH_HARD_LIMIT,
        min_message_id: int = 0,
    ) -> list[custom.Message]:
        limit = min(MESSAGE_FETCH_HARD_LIMIT, limit)

        logging.debug(
            "TelegramToRssClient.get_dialog_messages %s (%s) %s %s",
            dialog.name,
            dialog.id,
            limit,
            min_message_id,
        )

        messages: list[custom.Message] = await self._telethon.iter_messages(
            dialog, limit=limit, min_id=min_message_id
        ).collect()
        return messages

    @property
    def qr_code_url(self):
        return self._qr_code_url

    @property
    def user(self):
        return self._user


def telethon_dialog_id_to_tg_id(id: int):
    return resolve_id(id)[0]



from telegram_to_rss.client import TelegramToRssClient, custom, types
from telegram_to_rss.models import Feed, FeedEntry
from tortoise.expressions import Q
from tortoise.transactions import atomic
from pathlib import Path
import logging


class TelegramPoller:
    _client: TelegramToRssClient
    _message_limit: int
    _new_feed_limit: int
    _static_path: Path

    def __init__(
        self,
        client: TelegramToRssClient,
        message_limit: int,
        new_feed_limit: int,
        static_path: Path,
    ) -> None:
        self._client = client
        self._message_limit = message_limit
        self._new_feed_limit = new_feed_limit
        self._static_path = static_path

    async def fetch_dialogs(self):
        tg_dialogs = await self._client.list_dialogs()
        db_feeds = await Feed.all()

        tg_dialogs_ids = set([dialog.id for dialog in tg_dialogs])
        db_feeds_ids = set([feed.id for feed in db_feeds])

        feed_ids_to_delete = db_feeds_ids - tg_dialogs_ids
        feed_ids_to_create = tg_dialogs_ids - db_feeds_ids
        feed_ids_to_update = db_feeds_ids.intersection(tg_dialogs_ids)

        feeds_to_create = [
            dialog for dialog in tg_dialogs if dialog.id in feed_ids_to_create
        ]
        feeds_to_update = [
            dialog for dialog in tg_dialogs if dialog.id in feed_ids_to_update
        ]

        return (list(feed_ids_to_delete), feeds_to_create, feeds_to_update)

    async def bulk_delete_feeds(self, ids: list[int]):
        if len(ids) != 0:
            await Feed.filter(Q(id__in=list(ids))).delete()

    @atomic()
    async def create_feed(self, dialog: custom.Dialog):
        logging.debug("TelegramPoller.create_feed %s %s", dialog.name, dialog.id)

        feed = await Feed.create(id=dialog.id, name=dialog.name)

        logging.debug("TelegramPoller.create_feed -> get_dialog_messages")
        dialog_messages = await self._client.get_dialog_messages(
            dialog=dialog, limit=self._new_feed_limit
        )
        logging.debug("TelegramPoller.create_feed -> _process_new_dialog_messages")
        feed_entries = await self._process_new_dialog_messages(feed, dialog_messages)

        logging.debug("TelegramPoller.create_feed -> bulk_create")
        await FeedEntry.bulk_create(feed_entries)

    @atomic()
    async def update_feed(self, dialog: custom.Dialog):
        feed = await Feed.get(id=dialog.id)
        last_feed_entry = await FeedEntry.filter(feed=feed).order_by("-date").first()

        get_dialog_messages_args = {}
        if last_feed_entry:
            [_, tg_message_id] = parse_feed_entry_id(last_feed_entry.id)
            get_dialog_messages_args["min_message_id"] = tg_message_id
        else:
            get_dialog_messages_args["limit"] = self._new_feed_limit
            logging.warning(
                f"TelegramPoller.update_feed -> feed {feed.name} ({feed.id}) does not have associated feed entries"
            )

        new_dialog_messages = await self._client.get_dialog_messages(
            dialog=dialog, **get_dialog_messages_args
        )

        feed_entries = await self._process_new_dialog_messages(
            feed, new_dialog_messages
        )

        await FeedEntry.bulk_create(feed_entries)
        # Save even if unchanged to update date
        await feed.save()

        old_feed_entries = (
            await FeedEntry.filter(feed=feed)
            .order_by("-date")
            .limit(self._message_limit)
            .offset(self._message_limit)
        )

        await FeedEntry.filter(
            Q(id__in=[entry.id for entry in old_feed_entries])
        ).delete()

    async def _process_new_dialog_messages(
        self, feed: Feed, dialog_messages: list[custom.Message]
    ):
        filtered_dialog_messages: list[custom.Message] = []
        for dialog_message in dialog_messages:
            logging.debug(
                "TelegramPoller._process_new_dialog_messages -> processing message %s %s %s %s",
                dialog_message.id,
                dialog_message.grouped_id,
                dialog_message.photo is not None,
                dialog_message.text,
            )

            if dialog_message.text is None:
                continue

            dialog_message.downloaded_media = []

            if (
                dialog_message.grouped_id is None
                or len(filtered_dialog_messages) == 0
                or dialog_message.grouped_id != filtered_dialog_messages[-1].grouped_id
            ):
                filtered_dialog_messages.append(dialog_message)

            if (
                len(filtered_dialog_messages) != 0
                and dialog_message.grouped_id == filtered_dialog_messages[-1].grouped_id
                and len(dialog_message.text) > len(filtered_dialog_messages[-1].text)
            ):
                filtered_dialog_messages[-1].text = dialog_message.text

            last_processed_message = filtered_dialog_messages[-1]
            if dialog_message.photo:
                feed_entry_media_id = "{}-{}".format(
                    to_feed_entry_id(feed, dialog_message),
                    len(last_processed_message.downloaded_media),
                )
                media_path = self._static_path.joinpath(feed_entry_media_id)

                def progress_callback(current, total, media_path=media_path):
                    logging.debug(
                        "TelegramPoller._process_new_dialog_messages -> downloading media %s: %s out of %s",
                        media_path,
                        current,
                        total,
                    )

                res_path = await dialog_message.download_media(
                    file=media_path, progress_callback=progress_callback
                )
                last_processed_message.downloaded_media.append(Path(res_path).name)

        feed_entries: list[FeedEntry] = []
        for dialog_message in filtered_dialog_messages:
            feed_entry_id = to_feed_entry_id(feed, dialog_message)
            feed_entries.append(
                FeedEntry(
                    id=feed_entry_id,
                    feed=feed,
                    message=dialog_message.text,
                    date=dialog_message.date,
                    media=dialog_message.downloaded_media,
                    has_unsupported_media=dialog_message.media is not None
                    and not isinstance(dialog_message.media, types.MessageMediaPhoto),
                )
            )

        return feed_entries


def to_feed_entry_id(feed: Feed, dialog_message: custom.Message):
    return "{}--{}".format(feed.id, dialog_message.id)


def parse_feed_entry_id(id: str):
    [channel_id, message_id] = id.split("--")
    return (int(channel_id), int(message_id))


async def update_feeds_in_db(telegram_poller: TelegramPoller):
    logging.debug("update_feeds_in_db")

    [feed_ids_to_delete, feeds_to_create, feeds_to_update] = (
        await telegram_poller.fetch_dialogs()
    )
    logging.debug(
        "update_feeds_in_db -> fetched dialogs %s %s %s",
        feed_ids_to_delete,
        [dialog.id for dialog in feeds_to_create],
        [dialog.id for dialog in feeds_to_update],
    )

    await telegram_poller.bulk_delete_feeds(feed_ids_to_delete)
    logging.debug("update_feeds_in_db -> deleted feeds %s", feed_ids_to_delete)

    for feed_to_create in feeds_to_create:
        logging.debug(
            "update_feeds_in_db.create_feed %s %s",
            feed_to_create.id,
            feed_to_create.name,
        )
        await telegram_poller.create_feed(feed_to_create)
        logging.debug("update_feeds_in_db.create_feed -> done")

    for feed_to_update in feeds_to_update:
        logging.debug(
            "update_feeds_in_db.update_feed %s %s",
            feed_to_update.id,
            feed_to_update.name,
        )
        await telegram_poller.update_feed(feed_to_update)
        logging.debug("update_feeds_in_db.create_feed -> done")
